/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
** Copyright (c) 2001 Phil Thompson <phil@river-bank.demon.co.uk>
** Copyright (c) 2002 Riverbank Computing Limited <info@riverbankcomputing.co.uk>
**
** This file is part of Qt Designer.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "uic.h"
#include <qfile.h>
#include <qimage.h>
#include <qstringlist.h>
#include <qdatetime.h>
#include <qfileinfo.h>
#define NO_STATIC_COLORS
#include <globaldefs.h>
#include <qregexp.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

struct EmbedImage
{
    int width, height, depth;
    int numColors;
    QRgb* colorTable;
    QString name;
    QString cname;
    bool alpha;
};

static QString convertToCIdentifier( const char *s )
{
    QString r = s;
    int len = r.length();
    if ( len > 0 && !isalpha( (char)r[0].latin1() ) )
	r[0] = '_';
    for ( int i=1; i<len; i++ ) {
	if ( !isalnum( (char)r[i].latin1() ) )
	    r[i] = '_';
    }
    return r;
}


static void embedData( QTextStream& out, const uchar* input, int nbytes )
{
    static const char hexdigits[] = "0123456789abcdef";
    QString s;
    for ( int i=0; i<nbytes; i++ ) {
	if ( (i%16) == 0 ) {
	    if (i)
		s += '"';

	    s += " \\\n    \"";
	    out << (const char*)s;
	    s.truncate( 0 );
	}
	uint v = input[i];
	s += "\\x";
	s += hexdigits[(v >> 4) & 15];
	s += hexdigits[v & 15];
	if ( i == nbytes-1 )
	    s += "\"\n";
    }
    if ( s.length() )
	out << (const char*)s;
}

static void embedData( QTextStream& out, const QRgb* input, int n )
{
    out << hex;
    const QRgb *v = input;
    for ( int i=0; i<n; i++ ) {
	if ( (i%14) == 0  )
	    out << endl << "    ";
	out << "\\x";
	out << hex << *v++;
	if ( i < n-1 )
	    out << ',';
    }
    out << dec; // back to decimal mode
}

void Uic::embed( QTextStream& out, const char* project, const QStringList& images )
{

    QString cProject = convertToCIdentifier( project );

    QStringList::ConstIterator it;
    out << "# Image collection for project '" << project << "'." << endl;
    out << "#" << endl;
    out << "# Generated from reading image files: " << endl;
    for ( it = images.begin(); it != images.end(); ++it )
	out << "#      " << *it << endl;
    out << "#" << endl;
    out << "# Created: " << QDateTime::currentDateTime().toString() << endl;
    out << "#      by: The PyQt User Interface Compiler (pyuic)" << endl;
    out << "#" << endl;
    out << "# WARNING! All changes made in this file will be lost!" << endl;

    out << endl;
    out << endl;
    out << "from qt import QImage, QMimeSourceFactory" << endl;
    out << endl;
    out << endl;


    QPtrList<EmbedImage> list_image;
    int image_count = 0;
    for ( it = images.begin(); it != images.end(); ++it ) {
	QImage img;
	if ( !img.load( *it  ) ) {
	    fprintf( stderr, "pyuic: cannot load image file %s\n", (*it).latin1() );
	    continue;
	}
	EmbedImage *e = new EmbedImage;
	e->width = img.width();
	e->height = img.height();
	e->depth = img.depth();
	e->numColors = img.numColors();
	e->colorTable = new QRgb[e->numColors];
	e->alpha = img.hasAlphaBuffer();
	memcpy(e->colorTable, img.colorTable(), e->numColors*sizeof(QRgb));
	QFileInfo fi( *it );
	e->name = fi.fileName();
	e->cname = QString("image_%1").arg( image_count++);
	list_image.append( e );
	out << "# " << *it << endl;
	QString s;
	if ( e->depth == 32 ) {
	    out << s.sprintf( "%s_data =",
			      (const char *)e->cname );
	    embedData( out, img.bits(), e->width*e->height );
	} else {
	    if ( e->depth == 1 )
		img = img.convertBitOrder(QImage::LittleEndian);
	    out << s.sprintf( "%s_data =",
			      (const char *)e->cname );
	    embedData( out, img.bits(), img.numBytes() );
	}
	out << endl;
	if ( e->numColors ) {
	    out << s.sprintf( "%s_ctable = [",
			      (const char *)e->cname );
	    embedData( out, e->colorTable, e->numColors );
	    out << "\n]\n\n";
	}
    }

    if ( !list_image.isEmpty() ) {
	out << indent << "embed_image_vec = [\n";
	++indent;
	EmbedImage *e = list_image.first();
	while ( e ) {
	    out << indent << "["
		<< e->width << ","
		<< e->height << ","
		<< e->depth << ","
		<< e->cname << "_data,"
		<< e->numColors << ",";
	    if ( e->numColors )
		out << e->cname << "_ctable,";
	    else
		out << "None,";
	    if ( e->alpha )
		out << "1,";
	    else
		out << "0,";
	    out << "\"" << e->name << "\"],\n";
	    e = list_image.next();
	}
	--indent;
	out << indent << "]\n";

	out << "\n"
	    << indent << cProject << "image_dict = {}\n"
	    "\n"
	    "\n"
	    "def uic_findImage_" << cProject << "(name):\n";
	++indent;
	out << indent << "global " << cProject << "image_dict, embed_image_vec" << endl;
	out << endl;
	out << indent << "try:" << endl;
	++indent;
	out << indent << "return " << cProject << "image_dict[name]" << endl;
	--indent;
	out << indent << "except KeyError:" << endl;
	++indent;
	out << indent << "pass" << endl;
	--indent;
	out << endl;
	out << indent << "for (w,h,depth,data,nrCol,colTab,alpha,iname) in embed_image_vec:" << endl;
	++indent;
	out << indent << "if iname == name:" << endl;
	++indent;
	out << indent << "img = QImage(data,w,h,depth,colTab,nrCol,QImage.LittleEndian)" << endl;
	out << endl;
	out << indent << "if alpha:" << endl;
	++indent;
	out << indent << "img.setAlphaBuffer(1)" << endl;
	--indent;
	out << endl;
	out << indent << cProject << "image_dict[name] = img" << endl;
	out << indent << "return img" << endl;
	--indent;
	--indent;
	out << endl;
	out << indent << "return QImage()" << endl;
	--indent;
	out << endl;
	out << endl;

	out << indent << "class DesignerMimeSourceFactory_" << cProject << "(QMimeSourceFactory):" << endl;
	++indent;
	out << indent << "def data(self,abs_name):" << endl;
	++indent;
	out << indent << "img = uic_findImage_" << cProject << "(abs_name)" << endl;
	out << indent << "if img.isNull():" << endl;
	++indent;
	out << indent << "QMimeSourceFactory.removeFactory(self)" << endl;
	out << indent << "s = QMimeSourceFactory.defaultFactory().data(abs_name)" << endl;
	out << indent << "QMimeSourceFactory.addFactory(self)" << endl;
	out << indent << "return s" << endl;
	--indent;
	out << endl;
	out << indent << "QMimeSourceFactory.defaultFactory().setImage(abs_name,img)" << endl;
	out << indent << "return QMimeSourceFactory.defaultFactory().data(abs_name)" << endl;
	--indent;
	--indent;
	out << endl;
	out << endl;

	out << indent << "designerMimeSourceFactory = None" << endl;
	out << endl;

	out << indent << "def qInitImages_" << cProject << "():" << endl;
	++indent;
	out << indent << "global designerMimeSourceFactory" << endl;
	out << endl;
	out << indent << "if designerMimeSourceFactory is None:" << endl;
	++indent;
	out << indent << "designerMimeSourceFactory = DesignerMimeSourceFactory_" << cProject << "()" << endl;
	out << indent << "QMimeSourceFactory.defaultFactory().addFactory(designerMimeSourceFactory)" << endl;
	--indent;
	--indent;
	out << endl;
	out << endl;

	out << indent << "def qCleanupImages_" << cProject << "():" << endl;
	++indent;
	out << indent << "global " << cProject << "image_dict, designerMimeSourceFactory" << endl;
	out << endl;
	out << indent << cProject << "image_dict = {}" << endl;
	out << endl;
	out << indent << "if designerMimeSourceFactory is not None:" << endl;
	++indent;
	out << indent << "QMimeSourceFactory.defaultFactory().removeFactory(designerMimeSourceFactory)" << endl;
	out << indent << "designerMimeSourceFactory = None" << endl;
	--indent;
	--indent;
	out << endl;
	out << endl;

	out << indent << "class StaticInitImages_" << cProject << ":" << endl;
	++indent;
	out << indent << "def __init__(self):" << endl;
	++indent;
	out << indent << "qInitImages_" << cProject << "()" << endl;
	--indent;
	out << endl;
	out << indent << "def __del__(self):" << endl;
	++indent;
	out << indent << "qCleanupImages_" << cProject << "()" << endl;
	--indent;
	--indent;
	out << endl;
	out << endl;

	out << "staticImages = StaticInitImages_" << cProject << "()" << endl;
    }
}
